#!/usr/bin/perl 

use strict;
use warnings;
use WWW::Mechanize;
use File::Temp qw( tempfile tempdir );
use URI::Escape;
use Data::Dumper;
use File::Slurp qw( slurp );
use Data::RandomPerson;
use Math::BigInt;
use FindBin;

my $DEBUG = 1;
my $APP_PORT = 1337;
my $DATA_DIR = $FindBin::Bin;

my $task = $ARGV[0];

if ($task eq 'test') {
    my $ip = $ARGV[1];
    print "Testing for: $ip\n" if ($DEBUG);
    if (! test_cacert($ip)) {
        print "CA-Cert-Test failed\n" if ($DEBUG);
        exit 9;
    }
    my $p12_out = test_issue_p12($ip);
    if (ref $p12_out ne 'HASH') {
        print "PKCS#12-Issuance-Test failed\n" if ($DEBUG);
        exit 9;
    }
    if (! test_revocation(
            $ip, $p12_out->{SUBJECT}, $p12_out->{KEY}, $p12_out->{CERT}
        )) {
        print "Revocation test failed\n" if ($DEBUG);
        exit 9;
    }
    # TODO - test user key compromise flagging(?)
    # this is a functionality that is useful for the defending team,
    # so testing is is IMHO not really necessary ...
    
    print "all tests successful\n" if ($DEBUG);
    print STDERR "All tests successful";
    exit 0;
}

if ($task eq 'store') {
    my $ip     = $ARGV[1];
    my $flagid = $ARGV[2];
    my $flag   = $ARGV[3];

    # every ~ 30th time, register a new certificate or user
    # to make sure registration still works
    my $rand = int(rand(30));
    if ($rand == 1) {
        print "the random dice has decided to scrap our current certificate\n" if ($DEBUG);
        unlink "$DATA_DIR/$ip-subject.txt";
        unlink "$DATA_DIR/$ip-cert.txt";
    }
    if ($rand == 2) {
        print "the random dice has decided to scrap our current user\n" if ($DEBUG);
        unlink "$DATA_DIR/$ip-user.txt";
        unlink "$DATA_DIR/$ip-pass.txt";
    }
    if (! make_sure_cert_exists($ip)) {
        print "problem during make_sure_cert_exists\n" if ($DEBUG);
        # unlink files to make sure a new certificate is registered next time
        unlink "$DATA_DIR/$ip-subject.txt";
        unlink "$DATA_DIR/$ip-cert.txt";
        exit 21;
    }
    my $session_id = make_sure_user_exists($ip);
    if (! defined $session_id) {
        print "problem during make_sure_user_exists\n" if ($DEBUG);
        # unlink files to make sure a new user is registered next time
        unlink "$ip-user.txt";
        unlink "$ip-pass.txt";
        exit 21;
    }
    print "session id: $session_id\n" if ($DEBUG);

    my $username = slurp "$DATA_DIR/$ip-user.txt";
    my $password = slurp "$DATA_DIR/$ip-pass.txt";

    my $recipient = slurp "$DATA_DIR/$ip-subject.txt";
    my $message_digest = send_message($ip, $session_id, $recipient, $flag);
    if (! defined $message_digest) {
        print "error sending message\n" if ($DEBUG);
        exit 21;
    }
    my $level = slurp "$DATA_DIR/$ip-cert_level.txt";

    open my $FLAGS, '>>', "$DATA_DIR/$ip-flags.txt";
    print $FLAGS $flagid . ',' . $message_digest . ',' . $username . ',' . $password . ',' . $level . "\n";
    close $FLAGS;
    print "storing flag worked just fine\n" if ($DEBUG);

    print STDERR "Storing flag worked fine.";
    exit 0;
}

if ($task eq 'retrieve') {
    my $ip     = $ARGV[1];
    my $flagid = $ARGV[2];
    my $flag   = $ARGV[3];

    my $mech = WWW::Mechanize->new();

    if (! -r "$DATA_DIR/$ip-user.txt") {
        print "no username known\n" if ($DEBUG);
        log_error("no username known for retrieve");
        exit 21;
    }
    if (! -r "$DATA_DIR/$ip-pass.txt") {
        print "no password known\n" if ($DEBUG);
        log_error("no password known for retrieve");
        exit 21;
    }
    my @flags = slurp "$DATA_DIR/$ip-flags.txt";
    my ($flag_info) = grep { $_ =~ /^$flagid/ } @flags;
    if (! $flag_info) {
        print "flag id not found locally\n" if ($DEBUG);
        log_error("flag id not found locally ... - !?");
        exit 21;
    }
    chomp $flag_info;
    print "flag_info: $flag_info\n" if ($DEBUG);
    my ($flag_info_id, $message_digest, $username, $password, $cert_level)
        = split q{,}, $flag_info;

    my $session_id = login_user($ip, $username, $password);
    if (! defined $session_id) {
        print "login failed\n" if ($DEBUG);
        log_error("login failed.");
        exit 21;
    }

    # first way of retrieving a flag - sent messages page of the user,
    # password-based login
    $mech->get(base_uri($ip) . 'sent_messages.cgi?sessId=' . $session_id);
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }
    if ($mech->content !~ /$flag/) {
        print "flag not found on sent_messages page\n" if ($DEBUG);
        log_error("flag not found on sent messages page");
        exit 5;
    }

    # second way of retrieving a flag - sent messages page of the user,
    # challenge/response-based login
    $session_id = make_sure_user_key_exists($ip, $username);
    if (! defined $session_id) {
        print "challenge/response login failed\n" if ($DEBUG);
        log_error("login using challenge/response failed.");
        exit 21;
    }
    $mech->get(base_uri($ip) . 'sent_messages.cgi?sessId=' . $session_id);
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }
    if ($mech->content !~ /$flag/) {
        print "flag not found on sent_messages page\n" if ($DEBUG);
        log_error("flag not found on sent messages page");
        exit 5;
    }

    # third way of retrieving a flag - retrieve encrypted message and decrypt
    # using private key

    $mech->get(base_uri($ip) . 'messages.cgi');
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }
    if ($mech->status() != 200) {
        log_error("HTTP status != 200 (messages.cgi): " . $mech->status . "\n" . $mech->content());
        exit 21;
    }
    if ($mech->content() !~ m{ $message_digest }xms) {
        log_error("Message digest is not listed on messages page:" . $mech->content());
        exit 21;
    }
    $mech->get(base_uri($ip) . 'show_message.cgi?digest=' . uri_escape($message_digest));
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }
    if ($mech->status() != 200) {
        log_error("HTTP status != 200 (show_message.cgi): "
            . $mech->status . "\n" . $mech->content());
        exit 21;
    }
    if ($mech->content() !~ m{ x-pkcs7-mime }xms) {
        log_error("show_message.cgi result does not look like S/MIME: " . $mech->content());
        exit 21;
    }
    my ($SMIME, $smime_filename) = tempfile();
    print $SMIME $mech->content();
    close $SMIME;

    my $key_location = sprintf "%s/cert_keys/%02d/%02d.pem", $DATA_DIR, $cert_level, __key_id($ip);
    my $decoded = `/foo/bar/openssl/bin/openssl smime -decrypt -inkey $key_location -in $smime_filename`;
    print "decoded: $decoded\n" if ($DEBUG);
    if ($?) {
        log_error("decryption of message failed");
        exit 21;
    }
    if ($decoded ne $flag) {
        log_error("decoded message is not the expected flag: " . $decoded);
        exit 5;
    }

    unlink $smime_filename;

    print "all ways to retrieve flags were successful\n" if ($DEBUG);
    print STDERR "Retrieving flag worked fine.";
    exit 0;
}

sub send_message {
    my $ip         = shift;
    my $session_id = shift;
    my $recipient  = shift;
    my $message    = shift;

    my @subjects = slurp "$DATA_DIR/spam_subjects.txt";
    my $subject  = $subjects[int(rand(scalar @subjects))]; 
    chomp($subject);

    print "sending message\n" if ($DEBUG);
    print "to: $recipient\n" if ($DEBUG);
    print "subject: $subject\n" if ($DEBUG);
    print "message: $message\n" if ($DEBUG);

    my $mech = WWW::Mechanize->new();
    $mech->get(base_uri($ip) . 'sendmessage.cgi?session_id=' . uri_escape($session_id) . '&recipient=' . uri_escape($recipient) . '&subject=' . uri_escape($subject) . '&message=' . uri_escape($message));
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }
    if ($mech->status() != 200) {
        log_error("HTTP status != 200 (sendmessage.cgi): "
            . $mech->status . "\n" . $mech->content());
        return undef;
    }
    my ($message_digest) = ($mech->content =~ m{ The\ message\ digest\ is\ ([0-9a-f]+) \.}xms);
    if (! defined $message_digest) {
        log_error("Couldn't find message digest on sendmessage.cgi result: " . $mech->content());
    }
    return $message_digest;
}

sub __key_id {
    # extract key id from IP -> IP a.b.c.d -> key id = b

    my $ip = shift;
    my ($key_id) = ($ip =~ m{\A [0-9]+ \. ([0-9]+) \. [0-9]+ }xms);
    return $key_id;
}

sub make_sure_user_key_exists {
    my $ip = shift;

    my $level = slurp "$DATA_DIR/$ip-user_level.txt";

    my $key_location = sprintf "%s/user_keys/%02d/%02d.pem", $DATA_DIR, $level, __key_id($ip);
    my $key    = slurp $key_location;
    my $pubkey = `/foo/bar/openssl/bin/openssl rsa -in $key_location -pubout 2>/dev/null`;
    chomp $pubkey;
    print "pubkey: $pubkey\n" if ($DEBUG);

    my $username = slurp "$DATA_DIR/$ip-user.txt";
    my $password = slurp "$DATA_DIR/$ip-pass.txt";

    my $mech = WWW::Mechanize->new();
    $mech->get(base_uri($ip) . 'userinfo.cgi?username=' . uri_escape($username) . '&pubkey=1');
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }
    if ($mech->status() != 200) {
        log_error("HTTP status != 200 (userinfo.cgi, pubkey): "
            . $mech->status . "\n" . $mech->content());
        return undef;
    }
    if ($mech->content eq '') {
        if (! register_key($ip, $username, $password, $pubkey)) {
            log_error("Problem registering new key");
            return undef;
        }
        $mech->get(base_uri($ip) . 'userinfo.cgi?username=' . uri_escape($username) . '&pubkey=1');
        if ($mech->content =~ /Connection refused/) {
            log_error("HTTP connection refused"); exit 1;
        }
    }
    if ($pubkey ne $mech->content) {
        log_error("Public key is not the expected one, resetting it");
        print "mech content: $mech->content\n" if ($DEBUG);
        if (! register_key($ip, $username, $password, $pubkey)) {
            log_error("Problem updating key");
            return undef;
        }
        return undef;
    }

    print "trying to do c/r login with username: $username\n" if ($DEBUG);

    my $session_id = login_user_cr($ip, $username, $key);

    # special case key compromise ...
    if ($session_id eq 'keycompromise') {
        print "server claims key has been compromised, verify that\n" if ($DEBUG);
        $mech->get(base_uri($ip) . 'userinfo.cgi?username=' . uri_escape($username) . '&revocation=1');
        if ($mech->content =~ /Connection refused/) {
            log_error("HTTP connection refused"); exit 1;
        }
        if ($mech->status() != 200) {
            log_error("HTTP status != 200 (userinfo.cgi, revocation): "
                . $mech->status . "\n" . $mech->content());
            return undef;
        }
        my @primes  = split qq{\n}, $mech->content;
        my $modulus = `/foo/bar/openssl/bin/openssl rsa -in $key_location -noout -modulus`;
        chomp $modulus;
        $modulus =~ s/Modulus=//;

        print "primes: @primes\n" if ($DEBUG);
        print "modulus: $modulus\n" if ($DEBUG);
        
        my $modulus_bi = Math::BigInt->new('0x' . $modulus);
        my $prime_bi = Math::BigInt->new($primes[0]);
        $prime_bi->bmul(Math::BigInt->new($primes[1]));

        if ($prime_bi->bcmp($modulus_bi) != 0) {
            log_error("User key compromise info is incorrect:" . $mech->content);
            return undef;
        }

        print "key has been compromised, next level\n" if ($DEBUG);
        # we can not login because the key was compromised, increase the
        # level and register a new key for the user, login again
        $level++;
        print "level: $level\n" if ($DEBUG);
        open my $LEVEL, '>', "$DATA_DIR/$ip-user_level.txt";
        print $LEVEL $level;
        close $LEVEL;

        $key_location = sprintf "%s/user_keys/%02d/%02d.pem", $DATA_DIR, $level, __key_id($ip);
        $key    = slurp $key_location;
        $pubkey = `/foo/bar/openssl/bin/openssl rsa -in $key_location -pubout`;
        chomp $pubkey;
        if (! register_key($ip, $username, $password, $pubkey)) {
            log_error("Problem updating key after compromise");
            return undef;
        }
        $session_id = login_user_cr($ip, $username, $key);
    }

    return $session_id;
}

sub login_user_cr {
    my $ip       = shift;
    my $username = shift;
    my $key      = shift;

    my $mech = WWW::Mechanize->new();
    $mech->get(base_uri($ip) . 'generate_challenge.cgi?username=' . uri_escape($username));
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }
    if ($mech->status() != 200) {
        log_error("HTTP status != 200 (generate_challenge.cgi): "
            . $mech->status . "\n" . $mech->content());
        return undef;
    }
    my ($challenge) = ($mech->content() =~ m{ <input\ type=hidden\ name=challenge\ value=([0-9a-f:]+)>}xms);
    print "challenge: $challenge\n" if ($DEBUG);
    if (! defined $challenge) {
        log_error("Challenge not generated in a correct way: " . $mech->content());
    }
    my ($KEYH, $keyfile) = tempfile();
    print $KEYH $key;
    close $KEYH;

    my $signature = `$DATA_DIR/signchallenge.pl $challenge $keyfile`;

    print "signature: $signature\n" if ($DEBUG);

    $mech->get(base_uri($ip) . 'login_cr.cgi?username=' . uri_escape($username) . '&challenge=' . uri_escape($challenge) . '&signature=' . uri_escape($signature));
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }
    if ($mech->status() != 200) {
        log_error("HTTP status != 200 (login_cr.cgi): "
            . $mech->status . "\n" . $mech->content());
        return undef;
    }
    if ($mech->content =~ m{ Incorrect\ challenge\ signature }xms) {
        log_error("login_cr returned 'Incorrect challenge signature'");
        return undef;
    }
    if ($mech->content =~ m{ Key\ compromised }xms) {
        log_error("login_cr failed: key compromise");
        return "keycompromise";
    }

    my ($session_id) = ($mech->content =~ m{ sessId=([0-9a-f]+)"}xms);
    return $session_id;
}

sub register_key {
    my $ip       = shift;
    my $user     = shift;
    my $password = shift;
    my $key      = shift;

    my $session_id = login_user($ip, $user, $password);
    if (! defined $session_id) {
        log_error("Problem logging in to register key");
        return undef;
    }

    my $mech = WWW::Mechanize->new();
    $mech->get(base_uri($ip) . 'registerkey.cgi?session_id=' . $session_id . '&key=' . uri_escape($key));
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }

    if ($mech->status() != 200) {
        log_error("HTTP status != 200 (registerkey.cgi): "
            . $mech->status . "\n" . $mech->content());
        return undef;
    }
    if ($mech->content !~ m{Your\ public\ key\ has\ been\ updated\.}xms) {
        log_error("Incorrect registerkey.cgi response: " . $mech->content());
        return undef;
    }
    return 1;
}


sub make_sure_user_exists {
    my $ip = shift;

    if (! -e "$ip-user.txt") {
        print "no user file, register new user for IP\n" if ($DEBUG);
        if (! register_user($ip)) {
            return undef;
        }
    }

    my $username = slurp "$DATA_DIR/$ip-user.txt";
    my $password = slurp "$DATA_DIR/$ip-pass.txt";

    print "trying to login with username: $username\n" if ($DEBUG);
    print "and password: $password\n" if ($DEBUG);

    my $session_id = login_user($ip, $username, $password);

    return $session_id;
}

sub login_user {
    my $ip   = shift;
    my $user = shift;
    my $pass = shift;

    my $mech = WWW::Mechanize->new();
    $mech->get(base_uri($ip) . 'login.cgi?username=' . $user . '&password=' . $pass);
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }
    if ($mech->status() != 200) {
        log_error("HTTP status != 200 (login.cgi): "
            . $mech->status . "\n" . $mech->content());
        return undef;
    }
    my ($session_id) = ($mech->content =~ m{ sessId=([0-9a-f]+)"}xms);
    return $session_id;
}

sub register_user {
    my $ip = shift;

    if (! -e "$ip-user_level.txt") {
        open my $LEVEL, '>', "$DATA_DIR/$ip-user_level.txt";
        print $LEVEL 1;
        close $LEVEL;
    }
    my $level = slurp "$DATA_DIR/$ip-user_level.txt";

    my $mech = WWW::Mechanize->new();
    $mech->get(base_uri($ip) . 'register.html');
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }
    if ($mech->status() != 200) {
        log_error("HTTP status != 200 (register.html): "
            . $mech->status . "\n" . $mech->content());
        return 0;
    }
    my $drp      = Data::RandomPerson->new();
    my $p        = $drp->create();
    my $username = lc($p->{firstname}) . int(rand(1000));
    my $password = __rand_password();

    $mech->get(base_uri($ip) . 'register.cgi?username=' . uri_escape($username) . '&password=' . uri_escape($password));
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }
    if ($mech->status() != 200) {
        log_error("HTTP status != 200 (register.cgi): "
            . $mech->status . "\n" . $mech->content());
        return 0;
    }
    open my $USER, '>', "$DATA_DIR/$ip-user.txt";
    print $USER $username;
    close $USER;
    open my $PASS, '>', "$DATA_DIR/$ip-pass.txt";
    print $PASS $password;
    close $PASS;

    return 1;
}

sub make_sure_cert_exists {
    my $ip = shift;
    if (! -e "$DATA_DIR/$ip-subject.txt" || ! -e "$DATA_DIR/$ip-cert.txt") {
        print "no subject or cert file, register new certificate for IP\n" if ($DEBUG);
        if (! register_cert($ip)) {
            return 0;
        }
    }
    my $subject = slurp($DATA_DIR . '/' . $ip . '-subject.txt');
    my $cert    = slurp($DATA_DIR . '/' . $ip . '-cert.txt');
    print "retrieving cert for subject: $subject\n" if ($DEBUG);

    my $mech = WWW::Mechanize->new();
    $mech->get(base_uri($ip) . 'certinfo.cgi?subject=' . uri_escape($subject));
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }
    if ($mech->status() != 200) {
        log_error("HTTP status != 200 (certinfo.cgi): "
            . $mech->status . "\n" . $mech->content());
        return 0;
    }

    if ($mech->content !~ /-----BEGIN CERTIFICATE-----/) {

        log_error("certinfo.cgi output does not look like a cert: "
            . $mech->content);
        return 0;
    }
    
    if ($cert ne $mech->content) {
        log_error("certificate is not the one we have originally been issued : " . $mech->content);
        return 0;
    }

    # check certificate revocation, if valid, increase level and
    # register new one

    $mech->get(base_uri($ip) . 'show_revocation_info.cgi?subject=' . uri_escape($subject));
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }
    if ($mech->status() != 200) {
        log_error("HTTP status != 200 (show_revocation_info.cgi): "
            . $mech->status . "\n" . $mech->content());
        return 0;
    }
    if ($mech->content()) {
        my $smime_content = $mech->content();
        # retrieve CA certificate
        $mech->get(base_uri($ip) . 'get_ca_cert.cgi?format=PEM');
        if ($mech->content =~ /Connection refused/) {
            log_error("HTTP connection refused"); exit 1;
        }
        if ($mech->status() != 200) {
            log_error("HTTP status != 200 (get_ca_cert.cgi): "
                .$mech->status . "\n" . $mech->content());
            return 0;
        }
        if ($mech->content() !~ m{ \A -----BEGIN\ CERTIFICATE----- }xms) {
            log_error("CA certificate does not look like a certificate : "
                . $mech->content());
            return 0;
        }
        open my $CACERT, '>', "$DATA_DIR/$ip-cacert.txt";
        print $CACERT $mech->content;
        close $CACERT;

        my ($SMIME, $smime_filename) = tempfile();
        print $SMIME $smime_content;
        close $SMIME;

        print "smime_filename: $smime_filename\n" if ($DEBUG);
        # just checking for the CA certificate has the interesting
        # vulnerability of being able to sign this info using ANY
        # certificate issued by the CA - let's see if someone figures
        # this out :)
        my $signed = `/foo/bar/openssl/bin/openssl smime -verify -in $smime_filename -CAfile $DATA_DIR/$ip-cacert.txt 2>/dev/null`;
        if ($? != 0) {
            log_error("S/MIME verification on revocation info failed");
            return 0;
        }
        if ($signed ne $ip) {
            log_error("Signed information is not the IP" . $signed);
            return 0;
        }
        unlink $smime_filename;

        # update certificate level
        my $level = slurp "$DATA_DIR/$ip-cert_level.txt";
        $level++;
        open my $LEVEL, '>', "$DATA_DIR/$ip-cert_level.txt";
        print $LEVEL $level;
        close $LEVEL;

        if (! register_cert($ip)) {
            log_error("Problem registering new cert after old one was revoked");
            return 0;
        }
    }

    return 1;
}

sub register_cert {
    my $ip = shift;
    if (! -e "$ip-cert_level.txt") {
        open my $LEVEL, '>', "$DATA_DIR/$ip-cert_level.txt";
        print $LEVEL 1;
        close $LEVEL;
    }
    my $level = slurp "$DATA_DIR/$ip-cert_level.txt";
    
    my $mech = WWW::Mechanize->new();
    print "retrieving get_free_cert\n" if ($DEBUG);
    $mech->get(base_uri($ip) . 'get_free_cert.cgi');
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }
    my ($subject) = ($mech->content() =~ m{ name="subject"\ value="(.*?)" }xms);
    print "subject: $subject\n" if ($DEBUG);

    my $key_location = sprintf("%s/cert_keys/%02d/%02d.pem", $DATA_DIR, $level, __key_id($ip));

    my $drp  = Data::RandomPerson->new();
    my $p    = $drp->create();
    my $subj = $p->{firstname} . ' ' . $p->{lastname};

    open my $SUBJ, '>', "$DATA_DIR/$ip-subject.txt";
    print $SUBJ $subj;
    close $SUBJ;

    my ($REQ, $reqfile) = tempfile();
    if (! -e $key_location) {
        # this is because we have run out of levels
        # just generate a new random 1024bit key
        my $key_dir = sprintf("%s/cert_keys/%02d", $DATA_DIR, $level);
        mkdir $key_dir;
        `/foo/bar/openssl/bin/openssl genrsa -out $key_location 1024 2>/dev/null`; 
    }
    `/foo/bar/openssl/bin/openssl req -new -key $key_location -out $reqfile -subj '/CN=$subj'`;

    my $req = slurp $reqfile;
    print "request: $req\n" if ($DEBUG);

    $mech->get(base_uri($ip) . 'issue_cert_req.cgi?request=' . uri_escape($req));
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }
    if ($mech->status() != 200) {
        log_error("HTTP status != 200 (issue_cert_req.cgi): "
            . $mech->status . "\n" . $mech->content());
        return 0;
    }

    if ($mech->content !~ /-----BEGIN CERTIFICATE-----/) {
        log_error("issue_cert_req.cgi output does not look like a cert: "
            . $mech->content);
        return 0;
    }

    my $output = __parse_cert($mech->content);

    print "output: $output\n" if ($DEBUG);
    if ($output !~ /Subject: CN=$subj/) {
        log_error("issued certificate can not be parsed by OpenSSL: " . $output);
        return 0;
    }

    open my $CERT, '>', "$DATA_DIR/$ip-cert.txt";
    print $CERT $mech->content;
    close $CERT;

    return 1;
}

sub __parse_cert {
    my $cert = shift;

    my ($TEMP, $tempfile) = tempfile();
    print $TEMP $cert;
    close($TEMP);

    open my $OPENSSL, "/foo/bar/openssl/bin/openssl x509 -noout -text -in $tempfile 2>&1|";
    my $output = slurp($OPENSSL);
    close($OPENSSL);
    unlink($tempfile);

    return $output;
}

sub test_revocation {
    my $ip      = shift;
    my $subject = shift;
    my $key     = shift;
    my $cert    = shift;

    my $mech = WWW::Mechanize->new();
    $mech->get(base_uri($ip) . 'revoke.html');
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }
    if ($mech->status() != 200) {
        log_error("HTTP status != 200 (revoke.html): "
            . $mech->status . "\n" . $mech->content());
        return 0;
    }
    $mech->get(base_uri($ip) . 'revoke.cgi?subject=' . uri_escape($subject) . '&privkey=' . uri_escape($key) . '*submit=Revoke!');
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }
    if ($mech->status() != 200) {
        log_error("HTTP status != 200 (revoke.cgi): "
            . $mech->status . "\n" . $mech->content());
        return 0;
    }
    if ($mech->content !~ /has been revoked/) {
        log_error("Unexpected response after revocation: " . $mech->content());
        return 0;
    }
    $mech->get(base_uri($ip) . 'show_revocation_info.cgi?subject=' . uri_escape($subject));
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }
    if ($mech->status() != 200) {
        log_error("HTTP status != 200 (show_revocation_info.cgi): "
            . $mech->status . "\n" . $mech->content());
        return 0;
    }
    my $revocation_smime = $mech->content();
    if ($revocation_smime !~ /x-pkcs7-signature/) {
        log_error("Revocation info does not look like S/MIME: " . $revocation_smime);
        return 0;
    }
    # create CAcert file including EE + CA certificate
    my ($CERTS, $certsfile) = tempfile();
    print $CERTS $cert;
    close $CERTS;
    `cat $DATA_DIR/$ip-cacert.pem >> $certsfile`;

    my ($OUTHANDLE, $output) = tempfile;
    open my $SMIME, "|/foo/bar/openssl/bin/openssl smime -verify -CAfile $certsfile -out $output 2>/dev/null";
    print $SMIME $revocation_smime;
    close $SMIME;
    my $content = <$OUTHANDLE>;
    if ($? != 0) {
        log_error("Verification of revocation info failed, input: " . $revocation_smime);
        return 0;
    }
    chomp $content;
    if ($content ne $ip) {
        log_error("Signed revocation info is not the IP: " . $content);
        return 0;
    }
    return 1;
}

sub __rand_password {
    my $password = `/foo/bar/openssl/bin/openssl rand 128 -base64`;
    $password =~ s/=//g;
    $password =~ s/\///g;
    $password =~ s/\+//g;
    $password =~ s{\n}{}xmsg;
    return $password;
}

sub test_issue_p12 {
    my $ip = shift;

    my $password = __rand_password();
    # use random strings for passwords :-)
    print "password: $password\n" if ($DEBUG);

    my $mech = WWW::Mechanize->new();
    $mech->get(base_uri($ip) . 'get_free_cert.cgi');
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }
    my ($subject) = ($mech->content() =~ m{ name="subject"\ value="(.*?)" }xms);
    $mech->get(base_uri($ip) . 'issue_cert.cgi?subject='
        . uri_escape($subject) . '&password=' . uri_escape($password));
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }
    if ($mech->status() != 200) {
        log_error("HTTP status != 200: "
            . $mech->status . "\n" . $mech->content());
        return 0;
    }
    my ($TEMP, $tempfile) = tempfile();
    print $TEMP $mech->content();
    close($TEMP);

    my ($EECERT, $eecertfile) = tempfile();
    my ($EEKEY, $eekeyfile) = tempfile();

    $ENV{PASS} = $password;
    my $p12_output = `/foo/bar/openssl/bin/openssl pkcs12 -in $tempfile -passin env:PASS -nokeys -clcerts -out $eecertfile 2>/dev/null`;
    print '$?: ' . $? if ($DEBUG);
    if ($? != 0) {
        log_error("PKCS#12 decoding failed ($?): " . $p12_output);
        return 0;
    }

    my $eedump = `/foo/bar/openssl/bin/openssl x509 -in $eecertfile -noout -text`;
    if ($eedump !~ /Subject: CN=$subject, O=da-op3n010, C=DE/) {
        log_error("EE certificate has incorrect subject: " . $eedump);
        return 0;
    }

    my $verify = `/foo/bar/openssl/bin/openssl verify -CAfile $DATA_DIR/$ip-cacert.pem $eecertfile 2>/dev/null`;
    if ($verify !~ /: OK/) {
        log_error("CA signature verification error: " . $verify);
        return 0;
    }

    `/foo/bar/openssl/bin/openssl pkcs12 -in $tempfile -passin env:PASS -nodes -nocerts > $eekeyfile 2>/dev/null`;
    my $privkey_modulus = `/foo/bar/openssl/bin/openssl rsa -modulus -noout -in $eekeyfile`;
    my $cert_modulus = `/foo/bar/openssl/bin/openssl x509 -noout -modulus -in $eecertfile`;
    if ($cert_modulus ne $privkey_modulus) {
        log_error("Certificate and private key public modulus differ");
        return 0;
    }

    my $eekey = slurp($EEKEY);
    $eekey =~ s{.*-----BEGIN}{-----BEGIN}xms;
    my $eecert = slurp($EECERT);
    $eecert =~ s{.*-----BEGIN}{-----BEGIN}xms;
    print "eecert: $eecert\n" if ($DEBUG);
    unlink($tempfile);
    unlink($eecertfile);
    unlink($eekeyfile);

    return {
        SUBJECT => $subject,
        KEY     => $eekey,
        CERT    => $eecert,
        RESULT  => 1,
    };
}

sub test_cacert {
    my $ip = shift;

    my $mech = WWW::Mechanize->new();
    $mech->get(base_uri($ip) . 'get_ca_cert.cgi');
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }
    if ($mech->status() != 200) {
        log_error("HTTP status != 200 (PEM): "
            . $mech->status . "\n" . $mech->content());
        return 0;
    }
    if ($mech->content() !~ /-----BEGIN CERTIFICATE-----/) {
        log_error("certificate does not look like PEM: " . $mech->content());
        return 0;
    }
    open my $CERT, '>', "$DATA_DIR/$ip-cacert.pem";
    print $CERT $mech->content;
    close($CERT);

    my $output = __parse_cert($mech->content);

    if ($output !~ /Subject: /) {
        log_error("PEM certificate can not be parsed by OpenSSL: " . $output);
        return 0;
    }

    # DER format
    $mech->get(base_uri($ip) . 'get_ca_cert.cgi?format=DER');
    if ($mech->content =~ /Connection refused/) {
        log_error("HTTP connection refused"); exit 1;
    }
    # TODO - maybe log CA certs to see if teams changed them during the game?
    if ($mech->status() != 200) {
        log_error("HTTP status != 200 (DER): " 
            . $mech->status . "\n" . $mech->content());
        return 0;
    }
    my ($TEMP2, $tempfile2) = tempfile();
    print $TEMP2 $mech->content();
    close($TEMP2);

    open my $OPENSSL2, "/foo/bar/openssl/bin/openssl x509 -noout -text -inform DER -in $tempfile2 2>&1|";
    my $output2 = slurp($OPENSSL2);
    close($OPENSSL2);
    unlink($tempfile2);

    if ($output2 !~ /Subject: /) {
        log_error("DER certificate can not be parsed by OpenSSL: " . $output2);
        return 0;
    }

    return 1;
}

sub log_error {
    my $msg = shift;

    my ($package, $filename, $line, $subroutine, $hasargs,
        $wantarray, $evaltext, $is_require, $hints, $bitmask) = caller(1);
    if ($subroutine) {
        $subroutine =~ s/.*:://;
        $subroutine .= ': ';
    }
    else {
        $subroutine = "";
    }

    print STDERR substr($msg, 0, 255);
}

sub base_uri {
    my $ip = shift;
    return 'http://' . $ip . ':' . $APP_PORT . '/';
}
