#!/usr/bin/env python

import sys
import os
from pprint import pformat
from time import time

# Questions: hc@hcesperer.org

VERBOSE = False
clshow = ["\033[H\033[J"]

MSGS = {
    'intro': """<b>===== Testscript validator =====</b>
This script helps you test your testscripts.  Testing your testscripts
is a very important step when developing your service. Testscripts
should be adequately verbose, return the proper error codes, and be as
stable as possible. Also, they should not be vulnerable to attacks
from the network. <b>Especially beware of DoS-Attacks</b>, which get
nasty when exploited on the gameserver.\n""",
    'interactive': """Running in interactive mode. The purpose of this script is to test your
testscript. This is mostly an automatic process.""",
    'test': """Now testing your testscript. Please note that the test script's
descriptive message must be present at all times, giving a reasonable
explanation why the testscript returned correctly / why it failed""",
    'error_no_dmsg': """<b>No descriptive message</b>
Your testscript did not return a decent descriptive message. You have
to write an explanation why the testscript succeeded / failed to
stderr. This message must not be terminated with a newline character,
nor must it contain newlines itself.""",
    'error_nl': """Your descriptive message must not contain
newlines""",
    'error_dmsgtoolong': """Your descriptive message is too
long""",
    'error_dmsgtooshort': """Your descriptive message is too
short""",
    'error_bitfield': """Your error code is wrong. Please check
with the documentation.""",
    'error_nodebug': """Your testscript does not seem to produce appropriate
debug output to stdout. Debug output is very important for the ctf
organizers. Please provide it."""
}

ErrorCodes = {
    0: 'the connection to the service could not be established',
    1: 'the service returned a wrong flag',
    2: 'the service lacked some functionality at the time it was tested',
    3: 'the service did not respond in time',
    4: 'the service status is unknown',
    5: 'an error occured that is described in detail in the verbose error message',
    6: 'the service violated the protocol specification'
}

def cls(how=None):
    if how == None: how = clshow[0]
    print how

def pmsg(msg): print MSGS[msg].replace("<b>", "\033[1m").replace("</b>", "\033[0m")

def error(msg):
    cls()
    print "\033[1mYour testscript does not behave correctly.\033[0m"
    pmsg('error_%s' % msg)
    print "===== All tests aborted. Fix this first. ====="
    sys.exit(1)

def die(reason):
    sys.stderr.write("\033[1m%s\033[0m\n" % {
        'usage': 'USAGE: %s cmdline' % sys.argv[0],
        'bitfield': 'Bitfield return code type not supported ATM'
    }[reason])
    sys.exit(1)

def RunTS(cmdline, dontcheckdebug=False):
    if len(cmdline) < 1:
        pmsg('intro')
        die('usage')
    print '\033[1m===== BEGIN TESTING ====='
    print 'Cmdline: %s' % pformat(cmdline)
    print 'Running testscript...'
    p_in, p_out, p_err = os.popen3(cmdline, 't', 64 << 10)
    pid, res = os.wait()
    res = res >> 8
    success = res % 2
    bitfield = (res >> 1) % 2
    if bitfield: error('bitfield')
    code = res >> 2
    output = p_out.read()
    msg = p_err.read()
    print '== TESTSCRIPT OUTPUT ==\033[0m'
    print output
    print '\033[1m== RESULTS ==\033[0m'
    if VERBOSE: print 'Your testscript with with the pid %d returned %d.' % (pid, res)
    retcodetype = '%s' % {
        0: 'error/success code',
        1: 'bitfield'
    }[bitfield]
    if success: print """Your testscript reported \033[1mfailure\033[0m.
==> This does not mean that your testscript is faulty. We're testing your
    testscript here, not your service, remember? ;-) <==
The error code is %d, which \033[1m%s\033[0m.
Your testscript's verbose error message is: %s""" % \
        (code, (["means %s" % ErrorCodes[c] for c in [code] if c in \
        ErrorCodes] + ['is an unknown error code -- \033[1mERROR\033[0m!!'])[0],
        pformat(msg))

    else: print """Your testscript reported \033[1msuccessful operation\033[0m.
Verbose message: %s""" % msg

    if msg.strip() == '': error('no_dmsg')
    if msg.find("\n") != -1: error('nl')
    if len(msg) < {0: 5, 1: 16}[success]: error('dmsgtooshort')
    if len(msg) > 256: error('dmsgtoolong')
    if not dontcheckdebug:
        if len(output) < 16: error('nodebug')

    print '\033[0m'

    return success

def ReadLine(prompt=None):
    if prompt != None:
        sys.stdout.write("\033[1m%s>\033[0m " % prompt)
        sys.stdout.flush()
    return sys.stdin.readline()[:-1]

def dieif(what):
    if what:
        print """Could not store flag. Does not make sense to go on. As long as
the output of the testscript validator was correct, this means that
your service is broken/could not be reached, nothing more."""
        sys.exit(1)

def InteractiveTest():
    from sha import sha
    cls()
    pmsg('intro')
    pmsg('interactive')
    print "PRESS CTRL+C TO ABORT\n"
    print """Please enter the path to your testscript. You may also enter a
command line, for example 'python my_testscript.py'"""
    path = ReadLine("testscript")
    if len(path.strip()) == 0: exit(0)
    print "Please enter the IP where the appropriate service is listening:"
    s_ip = ReadLine("Service IP")
    if len(s_ip.strip()) == 0: exit(0)

    print "\nThanks. Press ENTER to store a flag."
    ReadLine()
    cls()
    flag = sha(str(time())).hexdigest()
    parms = [path, 'store', s_ip, sha(flag).hexdigest()[0:14], flag]
    dieif(RunTS(parms))
    print "\n\nFlag stored. Press ENTER to retrieve flag"
    ReadLine()
    cls()
    parms[1] = 'retrieve'
    res = RunTS(parms)
    if res: diecause("""Could not retrieve flag. Now that shouldn't happen.
Recode! Everything! From scratch! In another programming language, perhaps.""")
    print "\nNow trying to retrieve a nonexistant flag. Press ENTER to retrieve flag"
    ReadLine()
    cls()
    parms[3], parms[4] = 'fjiewaIJFEWAfewakjoaw', 'fkewaijiowafowaijfewaikoi'
    res = RunTS(parms)
    if not res: diecause("""Hmmm. Could retrieve flag. Check everything. Best, recode
it from scratch.""")
    print "\nThanks. Press ENTER to test additional service functionality."
    ReadLine()
    #cls()
    parms[1] = 'test'
    res = RunTS(parms[0:3], True)
    if res: diecause("That failed. But your testscript itself did fine.")
    print """

Congrats. It appears your testscript passed all tests. Permission
granted to svn commit it."""

if __name__ == '__main__':
    if len(sys.argv) > 2:
        clshow[0] = '' # no cls in non-interactive mode
        RunTS(list(sys.argv[1:]))
    else: InteractiveTest()
